{
***********************************************************
*     dlfunit.pas - a unit/DLL for manipulating DLF archives
*     Author: Prof A Olowofoyeku (The African Chief)
*     Licence : FREEWARE
*     Last amended: 26 August 2004
**************************************************************
}

{.$define __DLL__}
{.$define __debug__}

{$ifdef __DLL__}
LIBRARY dlfunit;
{$else}
UNIT dlfunit;
INTERFACE
{$endif}

{$I-,X+}

USES
Classes,
{$ifdef __GPC__}
 system,
{$endif}
{$ifndef VirtualPascal}
 {$ifndef linux}
   {$ifndef ver150}
   {$ifndef ver160}
      FileCtrl,
   {$endif}
   {$endif}
 {$endif}
{$endif}
searchr,
Sysutils;

{$ifdef __DLL__}
   {$undef __debug__}
   FUNCTION DLLVersion : Integer; STDCALL;
   BEGIN
      Result := 008;
   END;
{$endif}

// version details OF this code
CONST
ProgName = 'DLF tool';
ProgVer  = ' v0.8';
ProgAuthor = ' (c)2004 Prof. A Olowofoyeku (The African Chief)';
fOpenRead = fmOpenRead or fmShareDenyNone;

// slashes FOR Linux AND Dos / Windows
OS_Slash   = NativeOSDirSeparator;
OtherSlash = OtherOSDirSeparator;
WildCard = GlobalWildCard;
IsLinux = Unix_OS;

fNamelen = 30;
MaxFileNums = 2048;

TYPE
// data TYPE FOR FILE names IN DLF
FNameType = ARRAY [0 .. FNamelen - 1] OF char;

// data TYPE FOR directories IN DLF
DirType   = ARRAY [0 .. 14] OF char;

// our unsigned integer TYPE
Dlf_Word  = Cardinal;

// FOR dynamic buffers
CopyBufType = ARRAY [1.. ( 1024 * 1024 ) * 128] OF Byte; // max 128mb data size

// the DLF archive header
pDLF_HEADER = ^DLF_HEADER;
DLF_HEADER = PACKED RECORD
     FileCount,                // 4  = number OF files IN archive
     CompressionType,          // 8  = compression TYPE ?
     StartofFileData,          // 12 = offset FOR data
     HeaderEnd : Dlf_Word;     // 16 = END OF header ( zeroes )
END;

// FOR DLF archive FILE records
pDLF_FileRec = ^DLF_FileRec;
DLF_FileRec = PACKED RECORD
   FName  : FNameType;    // 30
   FSize  : DLF_Word;     // 34
   FDir   : DirType;      // 49
   FileID : DLF_Word;     // 53
END;

// TO store FILE offsets
FileOffsets = RECORD
  Start, Stop : DLF_Word;
END;

// full RECORD FOR files
pDLF_FileRecFull = ^DLF_FileRecFull;
DLF_FileRecFull = PACKED RECORD
   FName  : FNameType;    // 30
   FSize  : DLF_Word;     // 34
   FDir   : DirType;      // 49
   FileID : DLF_Word;     // 53
   Start,                 // 57
   Stop   : DLF_Word;     // 61
END;

// callback TYPE FOR listing DLF
TDLFListFunc = FUNCTION ( Head : pDLF_Header; FRrec : pDLF_FileRecFull )
: Integer {$ifdef Win32}STDCALL{$endif};

// callback TYPE FOR other processing OF DLF
TDLFArchiveFunc = FUNCTION ( FileName : pChar; Head : pDLF_HEADER; ID, Size, Copied : Longint )
: Integer {$ifdef Win32}STDCALL{$endif};

// ARRAY OF about 100kb size TO
// hold the FILE records ( up TO 2048 OF them )
pRecArrayType = ^RecArrayType;
RecArrayType = ARRAY [1..MaxFileNums] OF DLF_FileRec;

// ARRAY TO store FILE offsets
pOffSetArrayType = ^OffSetArrayType;
OffSetArrayType = ARRAY [1..MaxFileNums] OF FileOffSets;


// * * * * * * * * * * *  exported routines * * * * * * * * * * * * * * * * * * * //
// All these functions return a positive number IF successful,
// otherwise they return a negative error code

// create archive
FUNCTION CreateDLF ( ArchiveName, ArchiveFSpecs : pChar;
CallBack : TDLFArchiveFunc ) : Integer;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

// extract the contents OF the archive
FUNCTION ExtractDLF ( FileName, aTargetDir : pChar;
CallBack : TDLFArchiveFunc ) : Integer;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

// create archive
FUNCTION CreateDLFEx ( ArchiveName, ArchiveFSpecs : pChar ) : Integer;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

// extract the contents OF the archive
FUNCTION ExtractDLFEx ( FileName, aTargetDir : pChar ) : Integer;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}


// list the contents OF the archive
FUNCTION ListDLF ( FileName : pChar; CallBackFunc : TDLFListFunc ) : Integer;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

// obtain header information from DLF
FUNCTION ParseDLFArchive ( FileName : pChar;
                      VAR Head : DLF_Header;
                      VAR FRec : RecArrayType;
                      VAR Offsets : OffSetArrayType ) : Integer;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

// obtain header information from DLF
FUNCTION ParseDLFArchiveEx ( FileName : pChar;
                      Head : pChar;
                      FRec : pChar;
                      FRecLen : DLF_Word ) : Integer;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}


FUNCTION DLFTitle : pChar;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

FUNCTION DLFVersion : pChar;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

FUNCTION DLFAuthor : pChar;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

FUNCTION AddSlash ( CONST s : String ) : String;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

FUNCTION ToSlash ( CONST s : String ) : String;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

FUNCTION BackToFrontSlash ( CONST s : String ) : String;
{$ifdef __DLL__}STDCALL; FORWARD;{$endif}

{$ifndef __DLL__}
IMPLEMENTATION
  {$ifndef linux}
  uses windows;
  {$endif}
{$endif}

VAR
UsedWildCards : Boolean;
FileSpecs     : String;

FUNCTION AddSlash ( CONST s : String ) : String;
BEGIN
   Result := s;
   IF Result [Length ( s ) ] <> OS_Slash THEN
     Result := Result + OS_Slash;
END;

// change back TO front slashes OR vice versa AS appropriate
FUNCTION ToSlash ( CONST s : String ) : String;
VAR
i : DLF_Word;
BEGIN
   Result := s;
   FOR i := 1 TO Length ( s ) DO
     IF s [i] = OtherSlash THEN Result [i] := OS_Slash;
END;

// change back slashes TO front slashes
FUNCTION BackToFrontSlash ( CONST s : String ) : String;
VAR
i : DLF_Word;
BEGIN
   Result := s;
   FOR i := 1 TO Length ( s ) DO
     IF s [i] = '\' THEN Result [i] := '/';
END;

{$ifdef VirtualPascal}
FUNCTION DirectoryExists ( CONST s : String ) : Boolean;
VAR s1 : String;
BEGIN
  s1 := GetCurrentDir;
  ChDir ( s );
  Result := IOResult = 0;
  ChDir ( s1 );
END;
{$endif}

// READ AND parse the archive header AND data
// return information IN the various VAR parameters
//   DLF header information IS returned IN "Head"
//       Head.FileCount IS the total number OF files IN the DLF
//       Head.StartofFileData IS where the actual raw FILE data starts
//   FILE data IS returned IN "Frec"
//   FILE offsets are returned IN "Offsets"
FUNCTION ParseDLFArchive ( FileName  : pChar;
                      VAR Head : DLF_Header;
                      VAR FRec : RecArrayType;
                      VAR Offsets : OffSetArrayType ) : Integer;
VAR
f : FILE;
i, j : Cardinal;
BEGIN
  Result := - 1;
  FillChar ( Head, Sizeof ( Head ), #0 );
  FillChar ( Offsets, Sizeof ( Offsets ), #0 );
{$ifndef __DLL__} // FOR users OF DLL who can't create an array of this size
  FillChar ( Frec, Sizeof ( Frec ), #0 );
{$endif}

  IF Trim ( Strpas ( FileName ) ) = '' THEN Exit;
  Result := - 2;
  assign ( f, Trim ( Strpas ( FileName ) ) );
  i := FileMode;
  FileMode := fOpenRead; // open AS READ only
  Reset ( f, 1 );
  j := IOResult;
  FileMode := i;

  IF j <> 0 THEN exit;

  // READ header
  Result := - 3;
  BlockRead ( f, Head, sizeof ( Head ) );
  IF IOResult <> 0
  THEN BEGIN
     Close ( f );
     exit;
  END;

  // test header
  Result := - 4;
  WITH Head DO
  IF ( FileCount < 1 ) OR ( StartOfFileData < 1 )
  OR ( CompressionType <> 1 ) // some FILE error
  THEN BEGIN
     Close ( f );
     exit;
  END;

  // READ FILE data
  Result := - 5;
  BlockRead ( f, FRec, Sizeof ( DLF_FileRec ) * Head.FileCount );
  IF IOResult <> 0
  THEN BEGIN
     Close ( f );
     exit;
  END;
  Close ( f );

  // get the FILE offsets
  j := Head.StartOfFileData;
  FOR i := 1 TO Head.FileCount
  DO BEGIN
    WITH FRec [i]
    DO BEGIN
       IF i > 1 THEN Inc ( j, FRec [i - 1].FSize );
       WITH Offsets [i] DO BEGIN
           Start := j;
           Stop := Start + FSize;
       END;
    END;
  END;
  Result := Head.FileCount;
END; {* ParseDLFArchive *}


FUNCTION ParseDLFArchiveEx ( FileName : pChar;
                      Head : pChar;
                      FRec : pChar;
                      FRecLen : DLF_Word ) : Integer;
VAR
h : dlf_header;
i : Cardinal;
R : ^RecArrayType;
O : OffSetArrayType;
s : AnsiString;
BEGIN

  // get the header AND FILE information
  New ( r );
  Result := ParseDLFArchive ( FileName, h, r^, o );
  IF ( Result < 0 )
  THEN begin
     Dispose (r);
     Exit;
  end;

  With h
  do begin
     s := IntToStr (FileCount) + ', ' +
          IntToStr (CompressionType) + ', ' +
          IntToStr (StartOfFileData) + ', ' +
          IntToStr (HeaderEnd);
     StrpCopy (Head, s);
  end;

  Strpcopy ( FRec, '' );
  FOR i := 1 TO H.FileCount
  DO BEGIN
     WITH R^ [i]
     DO BEGIN
       s := Strpas (FName) + ', ' +
            IntToStr (FSize) + ', ' +
            Strpas (FDir) + ', ' +
            IntToStr (FileID) + ', ' +
            IntToStr (o [i].Start) + ', ' +
            IntToStr (o [i].Stop);
     END;

     // Safety margin: 100 character margin
     If Strlen ( FRec ) > ( FRecLen - 100 )
     then begin
        Result := i - 1; // return value must be less than filecount
        Break
     end
     else begin
        StrCat ( FRec, pChar ( s ) );
        StrCat ( FRec, '|' );
     end;
  END;
  StrCat ( FRec, #255 );
  Dispose (r);
END; {* ParseDLFArchiveEx *}

// list the contents OF the archive
FUNCTION ListDLF ( FileName : pChar; CallBackFunc : TDLFListFunc ) : Integer;
VAR
h : dlf_header;
i : Cardinal;
R : RecArrayType;
O : OffSetArrayType;
f : DLF_FileRecFull;
BEGIN

  // get the header AND FILE information
  Result := ParseDLFArchive ( FileName, h, r, o );
  IF ( Result < 0 ) THEN Exit;

  Result := - 200;
  IF NOT Assigned ( CallBackFunc ) THEN Exit;

  REsult := h.FileCount;

  // send header information only
  IF CallBackFunc ( pdlf_header ( @h ), NIL ) <> 0 THEN Exit;

  // display results
  FOR i := 1 TO h.FileCount
  DO BEGIN
    WITH R [i]
    DO BEGIN
       FillChar ( f, Sizeof ( f ), #0 );
       Strcopy ( f.FName, FName );
       f.FSize := FSize;
       Strcopy ( f.FDir, FDir );
       f.FileID := i;
       f.Start := o [i].Start;
       f.Stop := o [i].Stop;
       IF CallBackFunc ( pdlf_header ( @h ), pDLF_FileRecFull ( @f ) ) <> 0 THEN Break;
    END;
  END;
END; {* ListDLF *}


// extract the contents OF the archive
FUNCTION ExtractDLF ( FileName : pChar; aTargetDir : pChar;
CallBack : TDLFArchiveFunc ) : Integer;
VAR
h : dlf_header;
R : RecArrayType;
O : OffSetArrayType;
c, i, j, k : Longint;
TargetDir, Dest, s2 : AnsiString;
f1, f2 : FILE;
Buffer : ^CopyBufType;
BEGIN

  // get the header AND FILE information
  Result := ParseDLFArchive ( FileName, h, r, o );
  IF Result < 0 THEN Exit;

  TargetDir := Strpas ( aTargetDir );
  IF TargetDir = '' THEN TargetDir := GetCurrentDir;

  Result := - 100;
  Assign ( f1, Strpas ( FileName ) );
  i := FileMode;
  FileMode := fOpenRead;
  Reset ( f1, 1 );
  FileMode := i;

  IF IOResult <> 0 THEN Exit;

  // send notification OF header
  IF Assigned ( CallBack )
    THEN CallBack ( FileName, @h, - 1, - 1, - 1 );

  c := 0;
  FOR i := 1 TO h.FileCount
  DO BEGIN
    WITH R [i]
    DO BEGIN
       Dest := ToSlash ( Strpas ( FDir ) );
       IF Dest [1] IN ['\', '/'] THEN Delete ( Dest, 1, 1 );
       Dest := AddSlash ( TargetDir ) + Dest;
       IF NOT DirectoryExists ( Dest )
       THEN BEGIN
          Mkdir ( Dest );
          IF IOResult <> 0 THEN BEGIN
             {$ifdef __debug__}
             Writeln ( 'Error creating ', Dest );
             {$endif}
          END;
       END;

       IF DirectoryExists ( Dest )
       THEN BEGIN
          Seek ( f1, o [i].Start );
          s2 := AddSlash ( Dest ) + Strpas ( FName );
          Assign ( f2, s2 );
          ReWrite ( f2, 1 );
          IF IOResult = 0
          THEN BEGIN

              IF Assigned ( CallBack )
               THEN CallBack ( pChar ( s2 ), @h, i, FSize, - 100 );

              {$ifdef __debug__}
              Writeln ( 'Extracting ', s2, ' ... ' );
              {$endif}
              GetMem ( Buffer, FSize + 16 );
              BlockRead ( f1, Buffer^, FSize, j );
              BlockWrite ( f2, Buffer^, j, k );
               IF Assigned ( CallBack )
                  THEN CallBack ( pChar ( s2 ), @h, i, FSize, k );

              FreeMem ( Buffer, FSize + 16 );
              Close ( f2 );
              Inc ( c );
          END { IOResult = 0 }
          ELSE BEGIN
             {$ifdef __debug__}
             Writeln ( 'Cannot create FILE : ', s2 );
             {$endif}
          END;
       END { directoryexists }
       ELSE BEGIN
          {$ifdef __debug__}
          Writeln ( 'Cannot create target directory : ', Dest );
          {$endif}
       END;
    END;    { with }
  END;      { for }
  Close ( f1 );
  Result := c; // return number OF files successfully extracted
  // send notification OF header
  IF Assigned ( CallBack )
    THEN CallBack ( FileName, @h, - 2, - 2, - 2 );
END; {* ExtractDLF *}

FUNCTION CreateDLFEx ( ArchiveName, ArchiveFSpecs : pChar ) : Integer;
BEGIN
   Result := CreateDLF ( ArchiveName, ArchiveFSpecs, NIL );
END; {* CreateDLFEx *}

FUNCTION ExtractDLFEx ( FileName, aTargetDir : pChar ) : Integer;
BEGIN
   Result := ExtractDLF ( FileName, aTargetDir, NIL );
END; {* ExtractDLFEx *}

{ //////////////////////////////////////////// }
VAR
RootDir : String;
ArrCount : DLF_Word;
FArray : ^RecArrayType;
Flist  : TStringList;

FUNCTION HasWild ( CONST s : String ) : Boolean;
BEGIN
  HasWild :=
  ( s <> '' ) AND
  ( ( Pos ( '*', s ) <> 0 )  OR ( Pos ( '?', s ) <> 0 ) );
END;

FUNCTION SearchCallBack ( CONST afName : String; VAR Head :  TSearchRec ) :
Longint;
VAR
s, s1 : String;
b : Boolean;
BEGIN
    Result := 0;
    b := False;
    s := ExtractFileName ( aFName );
    {$ifdef __debug__}
    Writeln ('Name=', s, '| Fullname=', aFName,'|FSpecs=', FileSpecs);
    {$endif}

    { does the file match our search specs? }
    b := CompareFileNames ( s, FileSpecs ) = 0;
    If not b then begin
       IF UsedWildCards
       then begin
         b :=  ( MatchFileName ( FileSpecs, s ) )
         OR    ( MatchFileName ( ExtractFileName ( FileSpecs ), s ) )
       end // usedwildcards
       else begin

       end;
    end; // not b

    If ( b ) // we have a match
    and ( Head.Attr AND FaDirectory = 0 )// ignore directories
    THEN BEGIN
       Inc ( ArrCount );
       FList.Add ( aFName );
       WITH FArray^ [ArrCount]
       DO BEGIN
          s1 := ExtractFileDir ( aFName );
          IF Pos ( UpperCase ( RootDir ), UpperCase ( s1 ) ) = 1
           THEN Delete ( s1, 1, Length ( RootDir ) );

          StrpCopy ( FName, Head.Name );
          StrpCopy ( FDir, BackToFrontSlash ( s1 ) );
          IF StrLen ( FDir ) = 0 THEN StrCopy ( FDir, '\' );

          FSize := Head.Size;
          FileID := ArrCount;
       END;
    END;
END;

// create archive
FUNCTION CreateDLF ( ArchiveName, ArchiveFSpecs : pChar;
CallBack : TDLFArchiveFunc ) : Integer;
VAR
Header : DLF_HEADER;
ArcName, FSpecs,
Dir, Specs : String;
m,
i, j, k, l : DLF_Word;
ArcFile, f1 : FILE;
Buf : ARRAY [1..1024] OF Byte;
BEGIN
    Result := - 100;
    ArcName := Strpas ( ArchiveName );
    FSpecs := Strpas ( ArchiveFSpecs );
    Dir := ExtractFileDir ( FSpecs );
    Specs := ExtractFileName ( FSpecs );
    IF Dir = '' THEN Dir := GetCurrentDir;
    RootDir := Dir;

    {$ifdef __debug__}
    Writeln ('Archive=',Arcname,'|Specs=',FSpecs,'|Dir=',Dir);
    {$endif}

    ArrCount := 0;
    l := 0;
    New ( FArray );
    FillChar ( FArray^, Sizeof ( FArray^ ), #0 );
    FList := TStringList.Create;
    FileSpecs := FSpecs;
    UsedWildCards := HasWild ( FileSpecs );
    RecursiveSearch ( Dir, FileSpecs, SearchCallBack );
    {$ifdef __debug__}
    Writeln ('File Count=',ArrCount,'| Wildcards=', UsedWildCards);
    {$endif}
    IF ArrCount > 0
    THEN BEGIN
       WITH Header
       DO BEGIN
          FileCount := ArrCount;
          CompressionType := 1;
          StartofFileData := ( Sizeof ( DLF_FileRec ) * ArrCount ) + Sizeof ( Header );
          Dec ( StartofFileData, 4 ); {@@@@@@ needed for some reason }
          HeaderEnd := 0;
       END;

       // create archive FILE
       Result := - 300;
       Assign ( ArcFile, ArcName );
       ReWrite ( ArcFile, 1 );

       IF IOResult <> 0
       THEN BEGIN
         Dispose ( FArray );
         FList.Free;
         {$ifdef __debug__}
         Writeln ( 'Error creating FILE : ', ArcName );
         {$endif}
         Exit;
       END;

       {$ifdef __debug__}
       Writeln ( 'Creating DLF archive : ', ArcName, ' ...' );
       {$endif}

       // send notification OF header
       IF Assigned ( CallBack )
         THEN CallBack ( ArchiveName, @Header, - 1, - 1, - 1 );

       // WRITE header
       BlockWrite ( ArcFile, Header, Sizeof ( Header ), j );

       // WRITE FILE data
       FOR i := 1 TO ArrCount
       DO BEGIN
          BlockWrite ( ArcFile, FArray^ [i], Sizeof ( FArray^ [i] ), j );
       END;

       // archive the files
       Result := - 301;
       Seek ( ArcFile, Header.StartofFileData );
       FOR i := 0 TO Pred ( FList.Count )
       DO BEGIN
           Assign ( f1, FList [i] );
           m := FileMode;
           FileMode := fOpenRead;
           Reset ( f1, 1 );
           FileMode := m;
           IF IOResult = 0
           THEN BEGIN
              inc ( l );
              IF Assigned ( CallBack )
                THEN CallBack ( pChar ( FList [i] ), @Header, i + 1,
                  FArray^ [i + 1].FSize, - 100 );

              {$ifdef __debug__}
              Writeln ( 'Adding FILE : ', FList [i], ' ...' );
              {$endif}
              WHILE NOT Eof ( f1 )
              DO BEGIN
                 BlockRead ( f1, Buf, Sizeof ( Buf ), j );
                 BlockWrite ( ArcFile, Buf, j, k );
                 IF Assigned ( CallBack )
                   THEN CallBack ( pChar ( FList [i] ), @Header, i + 1,
                     FArray^ [i + 1].FSize, k );
              END;
              Close ( f1 );
           END { ioresult }
           ELSE BEGIN
             {$ifdef __debug__}
             Writeln ( 'Error creating FILE : ', FList [i], ' ...' );
             {$endif}
           END;
       END;
       // add dummy 4 bytes at the END TO pad size
       j := 192941350;
       BlockWrite ( ArcFile, j, Sizeof ( j ), k );

       Close ( ArcFile );

       // send notification OF header
       IF Assigned ( CallBack )
         THEN CallBack ( ArchiveName, @Header, - 2, - 2, - 2 );

       {$ifdef __debug__}
       Writeln ( 'Done.' );
       {$endif}
    END;
    Dispose ( FArray );
    FList.Free;
    Result := l;
END; {* CreateDLF *}

FUNCTION DLFTitle : pChar;
BEGIN
   Result := ProgName;
END;

FUNCTION DLFVersion : pChar;
BEGIN
   REsult := ProgVer;
END;

FUNCTION DLFAuthor : pChar;
BEGIN
   Result := ProgAuthor;
END;


///////////////////////////////
{$ifdef __DLL__}
EXPORTS
       DLLVersion      index 1 name 'DLLVersion',
       CreateDLF       index 2 name 'CreateDLF',
       ExtractDLF      index 3 name 'ExtractDLF',
       ListDLF         index 4 name 'ListDLF',
       ParseDLFArchive index 5 name 'ParseDLFArchive',
       DLFTitle        index 6 name 'DLFTitle',
       DLFVersion      index 7 name 'DLFVersion',
       DLFAuthor       index 8 name 'DLFAuthor',

       ParseDLFArchiveEx index 9 name 'ParseDLFArchiveEx',
       CreateDLFEx     index 10 name 'CreateDLFEx',
       ExtractDLFEx    index 11 name 'ExtractDLFEx';
BEGIN
{$endif}
END.

